#version 330 core

#include "settings.glsl"
#include "/lib/util/dither.glsl"

//in vec2 texcoord;
out vec4 FragColor;

#ifdef GL_ES
precision mediump float;
#endif

in vec2 vUv;
uniform sampler2D texture;

uniform sampler2D sceneTexture;
in vec2 fragTexCoord;

uniform float viewWidth, viewHeight, aspectRatio;
uniform vec3 cameraPosition, previousCameraPosition;
uniform mat4 gbufferPreviousProjection, gbufferProjectionInverse;
uniform mat4 gbufferModelView, gbufferPreviousModelView, gbufferModelViewInverse;
uniform sampler2D depthtex1;

ivec2 texelCoord = ivec2(gl_FragCoord.xy);
noperspective in vec2 texCoord;
uniform sampler2D colortex0;

vec3 fxaa(sampler2D tex, vec2 uv) {
    vec2 inverseScreenSize = 1.0 / vec2(textureSize(tex, 0));

    vec3 rgbNW = texture2D(tex, uv + vec2(-1.0, -1.0) * inverseScreenSize).xyz;
    vec3 rgbNE = texture2D(tex, uv + vec2(1.0, -1.0) * inverseScreenSize).xyz;
    vec3 rgbSW = texture2D(tex, uv + vec2(-1.0, 1.0) * inverseScreenSize).xyz;
    vec3 rgbSE = texture2D(tex, uv + vec2(1.0, 1.0) * inverseScreenSize).xyz;
    vec3 rgbM  = texture2D(tex, uv).xyz;

    vec3 luma = vec3(0.299, 0.587, 0.114);
    float lumaNW = dot(rgbNW, luma);
    float lumaNE = dot(rgbNE, luma);
    float lumaSW = dot(rgbSW, luma);
    float lumaSE = dot(rgbSE, luma);
    float lumaM  = dot(rgbM,  luma);

    float lumaMin = min(lumaM, min(min(lumaNW, lumaNE), min(lumaSW, lumaSE)));
    float lumaMax = max(lumaM, max(max(lumaNW, lumaNE), max(lumaSW, lumaSE)));

    vec2 dir;
    dir.x = -((lumaNW + lumaNE) - (lumaSW + lumaSE));
    dir.y =  ((lumaNW + lumaSW) - (lumaNE + lumaSE));

    float dirReduce = max((lumaNW + lumaNE + lumaSW + lumaSE) * (0.25 * 0.125), 0.0);
    float rcpDirMin = 1.0 / (min(abs(dir.x), abs(dir.y)) + dirReduce);
    dir = min(vec2(8.0, 8.0), max(vec2(-8.0, -8.0), dir * rcpDirMin)) * inverseScreenSize;

    vec3 rgbA = (1.0/2.0) * (
        texture2D(tex, uv + dir * (1.0 / 3.0 - 0.5)).xyz +
        texture2D(tex, uv + dir * (2.0 / 3.0 - 0.5)).xyz);

    vec3 rgbB = rgbA * (1.0/2.0) + (1.0/4.0) * (
        texture2D(tex, uv + dir * (0.0 / 3.0 - 0.5)).xyz +
        texture2D(tex, uv + dir * (3.0 / 3.0 - 0.5)).xyz);

    float lumaB = dot(rgbB, luma);

    if ((lumaB < lumaMin) || (lumaB > lumaMax)) {
        return rgbA;
    } else {
        return rgbB;
    }
}

vec3 applySharpening(vec3 color, float sharpness) {
    vec3 blurredColor = texture2D(texture, vUv).rgb;
    vec3 edge = blurredColor - color;
    return color + edge * sharpness;
}

vec3 MotionBlur(vec3 color, float z, float dither)
{
        if (z > 0.56) {
            float mbwg = 0.0;
            vec2 doublePixel = 2.0 / vec2(viewWidth, viewHeight);
            vec3 mblur = vec3(0.0);

            vec4 currentPosition = vec4(texCoord, z, 1.0) * 2.0 - 1.0;

            vec4 viewPos = gbufferProjectionInverse * currentPosition;
            viewPos = gbufferModelViewInverse * viewPos;
            viewPos /= viewPos.w;

            vec3 cameraOffset = cameraPosition - previousCameraPosition;

            vec4 previousPosition = viewPos + vec4(cameraOffset, 0.0);
            previousPosition = gbufferPreviousModelView * previousPosition;
            previousPosition = gbufferPreviousProjection * previousPosition;
            previousPosition /= previousPosition.w;

            vec2 velocity = (currentPosition - previousPosition).xy;
            velocity = velocity / (1.0 + length(velocity)) * MOTION_BLURRING_STRENGTH * 0.02;

            vec2 coord = texCoord - velocity * (3.5 + dither);
            for (int i = 0; i < 9; i++, coord += velocity) {
                vec2 coordb = clamp(coord, doublePixel, 1.0 - doublePixel);
                mblur += texture2DLod(colortex0, coordb, 0).rgb;
                mbwg += 1.0;
            }
            mblur /= mbwg;

            return mblur;
        } else return color;
}

vec3 calculateBloom(vec2 texCoord) {
    vec3 color = texture(sceneTexture, texCoord).rgb;
    float brightness = max(max(color.r, color.g), color.b);
    vec3 bloom = max(color - BLOOM_THRESHOLD, 0.0) * BLOOM_INTENSITY;
    return bloom;
}

void main() {
    #if ANTI_ALIASING == 1
    vec3 antiAliasedColor = fxaa(colortex0, vUv);
    vec3 finalColor = applySharpening(antiAliasedColor, SHARPNESS); // Adjust sharpness factor as needed
    #else
    vec3 finalColor = texture2D(colortex0, vUv.xy).rgb;
    #endif

    #if MOTION_BLUR == 1
    float z = texture2D(depthtex1, texCoord).x;
    float dither = Bayer64(gl_FragCoord.xy);
    vec3 motionBlurredColor = MotionBlur(finalColor, z, dither);
	#else
    vec3 motionBlurredColor = finalColor;
    #endif

    vec3 color = motionBlurredColor;


    // Apply saturation, brightness, and contrast
    float average = (color.r + color.g + color.b) / 3.0;
    color.rgb = mix(vec3(average), color.rgb, SATURATION);
    color.rgb += BRIGHTNESS;
    color.rgb = ((color.rgb - 0.5) * CONTRAST) + 0.5;

	// Apply bloom
	vec3 bloomColor = texture(sceneTexture, fragTexCoord).rgb;
	bloomColor = vec3(0.0);
	#if BLOOM == 1
	vec2 texelSize = 1.0 / textureSize(sceneTexture, 0);
    // Horizontal blur pass
    for (int i = -BLUR_RADIUS; i <= BLUR_RADIUS; ++i) {
        bloomColor += calculateBloom(fragTexCoord + vec2(float(i) * texelSize.x, 0.0));
    }
    // Vertical blur pass
    for (int i = -BLUR_RADIUS; i <= BLUR_RADIUS; ++i) {
        bloomColor += calculateBloom(fragTexCoord + vec2(0.0, float(i) * texelSize.y));
    }
	bloomColor /= float(2 * BLUR_RADIUS + 1);
    #endif

    /* DRAWBUFFERS:0 */
    gl_FragColor = vec4(color + bloomColor, 1.0);
}
